#
# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

import re
from typing import Any, Dict, List, Optional

import trafaret as t

from datarobot._compat import String
from datarobot.models.api_object import APIObject
from datarobot.utils.pagination import unpaginate

variable_trafaret = t.Dict({
    t.Key("name"): String(),
    t.Key("type"): String(),
    t.Key("description"): String(allow_blank=True),
}).ignore_extra("*")


prompt_template_version_trafaret = t.Dict({
    t.Key("id"): String(),
    t.Key("prompt_template_id"): String(),
    t.Key("prompt_text"): String(),
    t.Key("commit_comment"): String(allow_blank=True),
    t.Key("version"): t.Int,
    t.Key("variables"): t.List(variable_trafaret),
    t.Key("creation_date"): String(),
    t.Key("creation_user_id"): String(),
    t.Key("user_name"): String(allow_blank=True),
}).ignore_extra("*")


prompt_template_trafaret = t.Dict({
    t.Key("id"): String(),
    t.Key("name"): String(),
    t.Key("description"): String(allow_blank=True),
    t.Key("creation_date"): String(),
    t.Key("creation_user_id"): String(),
    t.Key("last_update_date"): String(),
    t.Key("last_update_user_id"): String(),
    t.Key("user_name"): String(allow_blank=True),
    t.Key("version_count"): t.Int,
}).ignore_extra("*")


class Variable(APIObject):
    """
    A variable used in prompt templates.

    Attributes
    ----------
    name : str
        Variable name.
    type : str
        Variable type (e.g., "str", "int").
    description : str
        Description of the variable.
    """

    _converter = variable_trafaret

    def __init__(self, name: str, type: str = "str", description: str = ""):
        self.name = name
        self.type = type
        self.description = description

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(name={self.name!r}, type={self.type!r})"

    def to_dict(self) -> Dict[str, str]:
        return {"name": self.name, "type": self.type, "description": self.description}


class PromptTemplateVersion(APIObject):
    """
    A specific version of a prompt template.

    Attributes
    ----------
    id : str
        Version ID.
    prompt_template_id : str
        ID of the parent prompt template.
    prompt_text : str
        The prompt text with variables in {{variable}} format.
    commit_comment : str
        Comment describing this version.
    version : int
        Version number.
    variables : List[Variable]
        List of variables used in the prompt.
    creation_date : str
        ISO-8601 formatted timestamp of when the version was created.
    creation_user_id : str
        ID of the user who created the version.
    user_name : str
        Name of the user who created the version.
    """

    _path = "api/v2/genai/promptTemplates"
    _converter = prompt_template_version_trafaret

    def __init__(
        self,
        id: str,
        prompt_template_id: Optional[str] = None,
        prompt_text: Optional[str] = None,
        commit_comment: Optional[str] = None,
        version: Optional[int] = None,
        variables: Optional[List[Dict[str, str]]] = None,
        creation_date: Optional[str] = None,
        creation_user_id: Optional[str] = None,
        user_name: Optional[str] = None,
    ):
        self.id = id
        self.prompt_template_id = prompt_template_id
        self.prompt_text = prompt_text
        self.commit_comment = commit_comment
        self.version = version
        self.variables = [Variable.from_server_data(v) for v in variables] if variables else []
        self.creation_date = creation_date
        self.creation_user_id = creation_user_id
        self.user_name = user_name

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(id={self.id!r}, "
            f"prompt_template_id={self.prompt_template_id!r}, version={self.version})"
        )

    @classmethod
    def get(cls, prompt_template_id: str, prompt_template_version_id: str) -> PromptTemplateVersion:
        """
        Get a specific prompt template version by ID.

        Parameters
        ----------
        prompt_template_id : str
            The ID of the prompt template.
        prompt_template_version_id : str
            The ID of the version to retrieve.

        Returns
        -------
        PromptTemplateVersion
            The requested prompt template version.

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> version = dr.genai.PromptTemplateVersion.get("prompt_template_id", "prompt_template_version_id")
        """
        url = f"{cls._client.domain}/{cls._path}/{prompt_template_id}/versions/{prompt_template_version_id}/"
        r_data = cls._client.get(url)
        return cls.from_server_data(r_data.json())

    @classmethod
    def create(
        cls,
        prompt_template_id: str,
        prompt_text: str,
        variables: Optional[List[Variable]] = None,
        commit_comment: str = "",
    ) -> PromptTemplateVersion:
        """
        Create a new version of a prompt template.

        Parameters
        ----------
        prompt_template_id : str
            The ID of the prompt template.
        prompt_text : str
            The prompt text with variables in {{variable}} format.
        variables : List[Variable], optional
            List of Variable objects defining the variables used in the prompt template.
        commit_comment : str, optional
            Comment describing this version.

        Returns
        -------
        PromptTemplateVersion
            The created version.

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> from datarobot.models.genai.prompt_template import Variable
            >>> version = dr.genai.PromptTemplateVersion.create(
            ...     prompt_template_id="507f1f77bcf86cd799439011",
            ...     prompt_text="Hello {{name}}, regarding {{issue}}",
            ...     variables=[
            ...         Variable(name="name", type="str", description="Customer name"),
            ...         Variable(name="issue", type="str", description="Issue type")
            ...     ],
            ...     commit_comment="Initial version"
            ... )
        """
        payload = {
            "promptText": prompt_text,
            "variables": [v.to_dict() for v in variables] if variables else [],
            "commitComment": commit_comment,
        }
        url = f"{cls._client.domain}/{cls._path}/{prompt_template_id}/versions/"
        response = cls._client.post(url, data=payload)
        return cls.from_server_data(response.json())

    @classmethod
    def list(
        cls,
        prompt_template_id: str,
    ) -> List[PromptTemplateVersion]:
        """
        List all versions of a prompt template.

        Parameters
        ----------
        prompt_template_id : str
            The ID of the prompt template.

        Returns
        -------
        List[PromptTemplateVersion]
            A list of versions for the template.

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> versions = dr.genai.PromptTemplateVersion.list(
            ...     prompt_template_id="507f1f77bcf86cd799439011"
            ... )
        """
        url = f"{cls._client.domain}/{cls._path}/{prompt_template_id}/versions/"
        data = unpaginate(url, None, cls._client)
        return [cls.from_server_data(item) for item in data]

    def render(self, variables: Optional[Dict[str, Any]] = None, **kwargs: Any) -> str:
        """
        Render the prompt text by substituting variables.

        Parameters
        ----------
        variables : Optional[Dict[str, Any]]
            Dictionary of variable names to values.
        **kwargs : Any
            Variable values as keyword arguments. These override values in the variables dict.

        Returns
        -------
        str
            Rendered prompt text with variables substituted.

        Raises
        ------
        ValueError
            If required variables are missing.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> version = dr.genai.PromptTemplateVersion.get("template_id", "version_id")
            >>> version.render(name="Alice", issue="billing")
            'Hello Alice, regarding billing...'
        """
        all_vars = {}
        if variables:
            all_vars.update(variables)
        all_vars.update(kwargs)

        if self.prompt_text is None:
            raise ValueError("Cannot render template: prompt_text is None")

        required_vars = {var.name for var in self.variables}
        provided_vars = set(all_vars.keys())
        missing = required_vars - provided_vars

        if missing:
            raise ValueError(f"Missing required variables: {', '.join(sorted(missing))}")

        rendered_text = self.prompt_text
        for var_name, var_value in all_vars.items():
            pattern = r"\{\{\s*" + re.escape(var_name) + r"\s*\}\}"
            rendered_text = re.sub(pattern, str(var_value), rendered_text)

        return rendered_text

    def to_fstring(self) -> str:
        """
        Convert the prompt text from {{variable}} format to Python f-string {variable} format.

        This method transforms template placeholders from double-brace format ({{variable}})
        to single-brace format ({variable}), making the template compatible with Python
        f-string syntax. Whitespace around variable names is automatically stripped.

        Returns
        -------
        str
            Prompt text with variables in f-string format.

        Raises
        ------
        ValueError
            If prompt_text is None.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> version = dr.genai.PromptTemplateVersion.get("template_id", "version_id")
            >>> # Original: "Hello {{name}}, regarding {{issue}}"
            >>> fstring_template = version.to_fstring()
            >>> # Result: "Hello {name}, regarding {issue}"
            >>> # Can now be used with f-string evaluation:
            >>> name = "Alice"
            >>> issue = "billing"
            >>> result = eval(f'f"{fstring_template}"')
        """
        if self.prompt_text is None:
            raise ValueError("Cannot convert template: prompt_text is None")

        import re

        fstring_text = re.sub(r'\{\{\s*(\w+)\s*\}\}', r'{\1}', self.prompt_text)

        return fstring_text


class PromptTemplate(APIObject):
    """
    A prompt template that can have multiple versions.

    Attributes
    ----------
    id : str
        Prompt template ID.
    name : str
        Prompt template name.
    description : str
        Description of the prompt template.
    creation_date : str
        ISO-8601 formatted timestamp of when the template was created.
    creation_user_id : str
        ID of the user who created the template.
    last_update_date : str
        ISO-8601 formatted timestamp of when the template was last updated.
    last_update_user_id : str
        ID of the user who last updated the template.
    user_name : str
        Name of the user who created the template.
    version_count : int
        Number of versions this template has.
    """

    _path = "api/v2/genai/promptTemplates"
    _converter = prompt_template_trafaret

    def __init__(
        self,
        id: str,
        name: Optional[str] = None,
        description: Optional[str] = None,
        creation_date: Optional[str] = None,
        creation_user_id: Optional[str] = None,
        last_update_date: Optional[str] = None,
        last_update_user_id: Optional[str] = None,
        user_name: Optional[str] = None,
        version_count: Optional[int] = None,
    ):
        self.id = id
        self.name = name
        self.description = description
        self.creation_date = creation_date
        self.creation_user_id = creation_user_id
        self.last_update_date = last_update_date
        self.last_update_user_id = last_update_user_id
        self.user_name = user_name
        self.version_count = version_count

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self.id!r}, name={self.name!r})"

    @classmethod
    def create(cls, name: str, description: str = "") -> PromptTemplate:
        """
        Create a new prompt template.

        Parameters
        ----------
        name : str
            Name of the prompt template.
        description : str, optional
            Description of the prompt template.

        Returns
        -------
        PromptTemplate
            The created prompt template.

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> template = dr.genai.PromptTemplate.create(
            ...     name="Customer Support",
            ...     description="Template for support responses"
            ... )
        """
        payload = {"name": name, "description": description}
        url = f"{cls._client.domain}/{cls._path}/"
        response = cls._client.post(url, data=payload)
        return cls.from_server_data(response.json())

    @classmethod
    def list(
        cls,
        sort: Optional[str] = None,
        search: Optional[str] = None,
    ) -> List[PromptTemplate]:
        """
        List all prompt templates available to the user.

        Parameters
        ----------
        sort : str, optional
            Property to sort by. Prefix with '-' for descending order.
            Example: '-creationDate'
        search : str, optional
            Filter templates by name containing this string.

        Returns
        -------
        List[PromptTemplate]
            A list of prompt templates.

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> templates = dr.genai.PromptTemplate.list(sort="-creationDate")
        """
        params: Dict[str, Any] = {}
        if sort is not None:
            params["sort"] = sort
        if search is not None:
            params["search"] = search

        url = f"{cls._client.domain}/{cls._path}/"
        data = unpaginate(url, params, cls._client)
        return [cls.from_server_data(item) for item in data]

    @classmethod
    def get(cls, prompt_template_id: str) -> PromptTemplate:
        """
        Get a prompt template by ID.

        Parameters
        ----------
        prompt_template_id : str
            The ID of the prompt template to retrieve.

        Returns
        -------
        PromptTemplate
            The requested prompt template.

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> template = dr.genai.PromptTemplate.get("507f1f77bcf86cd799439011")
        """
        url = f"{cls._client.domain}/{cls._path}/{prompt_template_id}/"
        r_data = cls._client.get(url)
        return cls.from_server_data(r_data.json())

    def create_version(
        self,
        prompt_text: str,
        variables: Optional[List[Variable]] = None,
        commit_comment: str = "",
    ) -> PromptTemplateVersion:
        """
        Create a new version of this prompt template.

        Parameters
        ----------
        prompt_text : str
            The prompt text with variables in {{variable}} format.
        variables : List[Variable], optional
            List of Variable objects defining the variables used in the prompt template.
            Defaults to None, which sends an empty list to the API.
        commit_comment : str, optional
            Comment describing this version.

        Returns
        -------
        PromptTemplateVersion
            The created version.

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.

        Examples
        --------
        .. code-block:: python

            >>> from datarobot.models.genai.prompt_template import Variable
            >>> template.create_version(
            ...     prompt_text="Hello {{name}}, regarding {{issue}}",
            ...     variables=[
            ...         Variable(name="name", type="str", description="Customer name"),
            ...         Variable(name="issue", type="str", description="Issue type")
            ...     ],
            ...     commit_comment="Initial version"
            ... )
        """
        return PromptTemplateVersion.create(self.id, prompt_text, variables, commit_comment)

    def list_versions(self) -> List[PromptTemplateVersion]:
        """
        List all versions of this prompt template.

        Returns
        -------
        List[PromptTemplateVersion]
            A list of versions for this template.

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.

        Examples
        --------
        .. code-block:: python

            >>> versions = template.list_versions()
        """
        return PromptTemplateVersion.list(self.id)

    def get_version(self, prompt_template_version_id: str) -> PromptTemplateVersion:
        """
        Get a specific version of this prompt template.

        Parameters
        ----------
        prompt_template_version_id : str
            The ID of the version to retrieve.

        Returns
        -------
        PromptTemplateVersion
            The requested version.

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> template = dr.genai.PromptTemplate.get("507f1f77bcf86cd799439011")
            >>> version = template.get_version("507f1f77bcf86cd799439012")
        """
        return PromptTemplateVersion.get(self.id, prompt_template_version_id)

    def get_latest_version(self) -> Optional[PromptTemplateVersion]:
        """
        Get the latest version of this prompt template.

        Returns
        -------
        Optional[PromptTemplateVersion]
            The latest version, or None if no versions exist.

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> template = dr.genai.PromptTemplate.get("507f1f77bcf86cd799439011")
            >>> latest = template.get_latest_version()
            >>> if latest:
            ...     print(f"Version {latest.version}")
        """
        versions = self.list_versions()
        if not versions:
            return None
        return max(versions, key=lambda v: v.version)  # type: ignore[arg-type, return-value]



#
# Copyright 2021-2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, List, Optional, Type, TypeVar, cast

import trafaret as t

from datarobot._compat import Int, String
from datarobot.models.api_object import APIObject
from datarobot.utils import from_api, parse_time, to_api

TFeaturelist = TypeVar("TFeaturelist", bound="Featurelist")
TModelingFeaturelist = TypeVar("TModelingFeaturelist", bound="ModelingFeaturelist")
TDatasetFeaturelist = TypeVar("TDatasetFeaturelist", bound="DatasetFeaturelist")

if TYPE_CHECKING:
    from datarobot._compat import TypedDict
    from datarobot.models.api_object import ServerDataDictType
    from datarobot.models.project import Project

    class DeleteFeatureListResult(TypedDict):
        dry_run: bool
        can_delete: bool
        deletion_blocked_reason: str
        num_affected_models: int
        num_affected_jobs: int


class BaseFeaturelist(APIObject):  # pylint: disable=missing-class-docstring
    _base_path: str  # path, to be overridden in inheriting classes
    # nulls are to account for the partial featurelist created inside models
    _converter = t.Dict({
        t.Key("id"): t.Or(String, t.Null),
        t.Key("name"): t.Or(String, t.Null),
        t.Key("features"): t.Or(t.List(String), t.Null),
        t.Key("project_id"): t.Or(String(), t.Null),
        t.Key("created"): t.Or(t.Null, t.Call(parse_time)),
        t.Key("is_user_created"): t.Or(t.Bool, t.Null),
        t.Key("num_models"): t.Or(Int, t.Null),
        t.Key("description"): String(allow_blank=True),
    }).ignore_extra("*")

    def __init__(
        self,
        id: Optional[str] = None,
        name: Optional[str] = None,
        features: Optional[List[str]] = None,
        project_id: Optional[str] = None,
        created: Optional[datetime] = None,
        is_user_created: Optional[bool] = None,
        num_models: Optional[int] = None,
        description: Optional[str] = None,
    ) -> None:
        self.id = id
        self.project_id = project_id
        self.name = name
        self.features = features
        self.created = created
        self.is_user_created = is_user_created
        self.num_models = num_models
        self.description = description

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.name})"

    def _make_url(self) -> str:
        return f"{self._base_path.format(self.project_id)}{self.id}/"

    def update(self, name: Optional[str] = None, description: Optional[str] = None) -> None:
        """Update the name or description of an existing featurelist

        Note that only user-created featurelists can be renamed, and that names must not
        conflict with names used by other featurelists.

        Parameters
        ----------
        name : Optional[str]
            the new name for the featurelist
        description : Optional[str]
            the new description for the featurelist
        """
        data = {}
        if name is not None:
            data["name"] = name
        if description is not None:
            data["description"] = description
        self._client.patch(self._make_url(), data=data)
        if name is not None:
            self.name = name
        if description is not None:
            self.description = description

    def delete(self, dry_run: bool = False, delete_dependencies: bool = False) -> DeleteFeatureListResult:
        """Delete a featurelist, and any models and jobs using it

        All models using a featurelist, whether as the training featurelist or as a monotonic
        constraint featurelist, will also be deleted when the deletion is executed and any queued or
        running jobs using it will be cancelled. Similarly, predictions made on these models will
        also be deleted. All the entities that are to be deleted with a featurelist are described
        as "dependencies" of it.  To preview the results of deleting a featurelist, call delete
        with `dry_run=True`

        When deleting a featurelist with dependencies, users must specify `delete_dependencies=True`
        to confirm they want to delete the featurelist and all its dependencies. Without that
        option, only featurelists with no dependencies may be successfully deleted and others will
        error.

        Featurelists configured into the project as a default featurelist or as a default monotonic
        constraint featurelist cannot be deleted.

        Featurelists used in a model deployment cannot be deleted until the model deployment is
        deleted.

        Parameters
        ----------
        dry_run : Optional[bool]
            specify True to preview the result of deleting the featurelist, instead of actually
            deleting it.
        delete_dependencies : Optional[bool]
            specify True to successfully delete featurelists with dependencies; if left
            False by default, featurelists without dependencies can be successfully deleted and
            those with dependencies will error upon attempting to delete them.

        Returns
        -------
        result : dict
            A dictionary describing the result of deleting the featurelist, with the following keys
                - dry_run : bool, whether the deletion was a dry run or an actual deletion
                - can_delete : bool, whether the featurelist can actually be deleted
                - deletion_blocked_reason : str, why the featurelist can't be deleted (if it can't)
                - num_affected_models : int, the number of models using this featurelist
                - num_affected_jobs : int, the number of jobs using this featurelist
        """
        result = self._client.delete(
            self._make_url(),
            params=to_api({"dry_run": dry_run, "delete_dependencies": delete_dependencies}),
        )
        checker = t.Dict({
            t.Key("dry_run"): t.Bool(),
            t.Key("can_delete"): t.Bool(),
            t.Key("deletion_blocked_reason"): String(allow_blank=True),
            t.Key("num_affected_models"): Int(),
            t.Key("num_affected_jobs"): Int(),
        }).ignore_extra("*")
        return cast("DeleteFeatureListResult", checker.check(from_api(result.json())))


class Featurelist(BaseFeaturelist):
    """A set of features used in modeling

    Attributes
    ----------
    id : str
        the id of the featurelist
    name : str
        the name of the featurelist
    features : List[str]
        the names of all the Features in the featurelist
    project_id : str
        the project the featurelist belongs to
    created : datetime.datetime
        (New in version v2.13) when the featurelist was created
    is_user_created : bool
        (New in version v2.13) whether the featurelist was created by a user or by DataRobot
        automation
    num_models : int
        (New in version v2.13) the number of models currently using this featurelist.  A model is
        considered to use a featurelist if it is used to train the model or as a monotonic
        constraint featurelist, or if the model is a blender with at least one component model
        using the featurelist.
    description : str
        (New in version v2.13) the description of the featurelist.  Can be updated by the user
        and may be supplied by default for DataRobot-created featurelists.
    """

    _base_path = "projects/{}/featurelists/"
    _project: Project

    def __init__(
        self,
        id: Optional[str] = None,
        name: Optional[str] = None,
        features: Optional[List[str]] = None,
        project_id: Optional[str] = None,
        created: Optional[datetime] = None,
        is_user_created: Optional[bool] = None,
        num_models: Optional[int] = None,
        description: Optional[str] = None,
    ) -> None:
        super().__init__(
            id=id,
            name=name,
            features=features,
            project_id=project_id,
            created=created,
            is_user_created=is_user_created,
            num_models=num_models,
            description=description,
        )
        self._set_up(self.project_id)

    @property
    def project(self) -> Project:
        return self._project

    def _set_up(self, project_id: Optional[str]) -> None:
        from datarobot.models.project import (  # pylint: disable=import-outside-toplevel,cyclic-import
            Project,
        )

        self._project = Project(project_id)

    @classmethod
    def from_data(cls: Type[TFeaturelist], data: ServerDataDictType) -> TFeaturelist:  # type: ignore[override]
        """Overrides the parent method to ensure description is always populated

        Parameters
        ----------
        data : dict
            the data from the server, having gone through processing
        """
        # these keys are sometimes missing b/c dummy featurelists are instantiated inside models
        # in _normal_ usage these should all be supplied
        for key in [
            "features",
            "num_models",
            "is_user_created",
            "created",
            "id",
            "name",
            "project_id",
        ]:
            data.setdefault(key, None)
        if "description" not in data:
            data["description"] = ""
        return super().from_data(data)

    @classmethod
    def get(cls: Type[TFeaturelist], project_id: str, featurelist_id: str) -> TFeaturelist:
        """Retrieve a known feature list

        Parameters
        ----------
        project_id : str
            The id of the project the featurelist is associated with
        featurelist_id : str
            The ID of the featurelist to retrieve

        Returns
        -------
        featurelist : Featurelist
            The queried instance

        Raises
        ------
        ValueError
            passed ``project_id`` parameter value is of not supported type
        """
        from datarobot.models.project import (  # pylint: disable=import-outside-toplevel,cyclic-import
            Project,
        )

        if isinstance(project_id, Project):  # type: ignore[unreachable]
            project_id = project_id.id  # type: ignore[unreachable]
        elif isinstance(project_id, str):
            pass
        else:
            raise ValueError("Project arg must be Project instance or str")

        url = f"{cls._base_path.format(project_id)}{featurelist_id}/"
        return cls.from_location(url)


class ModelingFeaturelist(BaseFeaturelist):
    """A set of features that can be used to build a model

    In time series projects, a new set of modeling features is created after setting the
    partitioning options.  These features are automatically derived from those in the project's
    dataset and are the features used for modeling.  Modeling features are only accessible once
    the target and partitioning options have been set.  In projects that don't use time series
    modeling, once the target has been set, ModelingFeaturelists and Featurelists will behave
    the same.

    For more information about input and modeling features, see the
    :ref:`time series documentation<input-vs-modeling>`.

    Attributes
    ----------
    id : str
        the id of the modeling featurelist
    project_id : str
        the id of the project the modeling featurelist belongs to
    name : str
        the name of the modeling featurelist
    features : List[str]
        a list of the names of features included in this modeling featurelist
    created : datetime.datetime
        (New in version v2.13) when the featurelist was created
    is_user_created : bool
        (New in version v2.13) whether the featurelist was created by a user or by DataRobot
        automation
    num_models : int
        (New in version v2.13) the number of models currently using this featurelist.  A model is
        considered to use a featurelist if it is used to train the model or as a monotonic
        constraint featurelist, or if the model is a blender with at least one component model
        using the featurelist.
    description : str
        (New in version v2.13) the description of the featurelist.  Can be updated by the user
        and may be supplied by default for DataRobot-created featurelists.
    """

    _base_path = "projects/{}/modelingFeaturelists/"

    @classmethod
    def get(cls: Type[TModelingFeaturelist], project_id: str, featurelist_id: str) -> TModelingFeaturelist:
        """Retrieve a modeling featurelist

        Modeling featurelists can only be retrieved once the target and partitioning options have
        been set.

        Parameters
        ----------
        project_id : str
            the id of the project the modeling featurelist belongs to
        featurelist_id : str
             the id of the modeling featurelist to retrieve

        Returns
        -------
        featurelist : ModelingFeaturelist
            the specified featurelist
        """
        url = f"{cls._base_path.format(project_id)}{featurelist_id}/"
        return cls.from_location(url)


class DatasetFeaturelist(APIObject):
    """A set of features attached to a dataset in the AI Catalog

    Attributes
    ----------
    id : str
        the id of the dataset featurelist
    dataset_id : str
        the id of the dataset the featurelist belongs to
    dataset_version_id: Optional[str]
        the version id of the dataset this featurelist belongs to
    name : str
        the name of the dataset featurelist
    features : List[str]
        a list of the names of features included in this dataset featurelist
    creation_date : datetime.datetime
        when the featurelist was created
    created_by : str
        the user name of the user who created this featurelist
    user_created : bool
        whether the featurelist was created by a user or by DataRobot automation
    description : Optional[str]
        the description of the featurelist. Only present on DataRobot-created featurelists.
    """

    _base_path = "datasets/{}/featurelists/"
    _converter = t.Dict({
        t.Key("id"): t.Or(String, t.Null),
        t.Key("name"): t.Or(String, t.Null),
        t.Key("features"): t.Or(t.List(String), t.Null),
        t.Key("dataset_id"): t.Or(String, t.Null),
        t.Key("dataset_version_id", optional=True): t.Or(String, t.Null),
        t.Key("creation_date"): t.Or(t.Null, t.Call(parse_time)),
        t.Key("created_by"): t.Or(String, t.Null),
        t.Key("user_created"): t.Or(t.Bool, t.Null),
        t.Key("description", optional=True): String(allow_blank=True),
    }).ignore_extra("*")

    def __init__(
        self,
        id: Optional[str] = None,
        name: Optional[str] = None,
        features: Optional[List[str]] = None,
        dataset_id: Optional[str] = None,
        dataset_version_id: Optional[str] = None,
        creation_date: Optional[datetime] = None,
        created_by: Optional[str] = None,
        user_created: Optional[bool] = None,
        description: Optional[str] = None,
    ) -> None:
        self.id = id
        self.name = name
        self.features = features
        self.dataset_id = dataset_id
        self.dataset_version_id = dataset_version_id
        self.creation_date = creation_date
        self.created_by = created_by
        self.user_created = user_created
        self.description = description

    def _make_url(self) -> str:
        return f"{self._base_path.format(self.dataset_id)}{self.id}/"

    @classmethod
    def get(cls: Type[TDatasetFeaturelist], dataset_id: str, featurelist_id: str) -> TDatasetFeaturelist:
        """Retrieve a dataset featurelist

        Parameters
        ----------
        dataset_id : str
            the id of the dataset the featurelist belongs to
        featurelist_id : str
             the id of the dataset featurelist to retrieve

        Returns
        -------
        featurelist : DatasetFeatureList
            the specified featurelist
        """
        url = f"{cls._base_path.format(dataset_id)}{featurelist_id}/"
        return cls.from_location(url)

    def delete(self) -> None:
        """Delete a dataset featurelist

        Featurelists configured into the dataset as a default featurelist cannot be deleted.
        """
        self._client.delete(self._make_url())

    def update(self, name: Optional[str] = None) -> None:
        """Update the name of an existing featurelist

        Note that only user-created featurelists can be renamed, and that names must not
        conflict with names used by other featurelists.

        Parameters
        ----------
        name : Optional[str]
            the new name for the featurelist
        """
        data = {}
        if name is not None:
            data["name"] = name
        self._client.patch(self._make_url(), data=data)
        if name is not None:
            self.name = name

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.name})"



#
# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import List, Optional

import trafaret as t

from datarobot._compat import TypedDict
from datarobot.models.api_object import APIObject
from datarobot.models.notebooks.enums import ImageLanguage

image_trafaret = t.Dict({
    t.Key("id"): t.String,
    t.Key("name"): t.String,
    t.Key("default"): t.Bool,
    t.Key("description"): t.String,
    t.Key("environment_id"): t.String,
    t.Key("gpu_optimized"): t.Bool,
    t.Key("language"): t.Enum(*list(ImageLanguage)),
    t.Key("language_version"): t.String,
    t.Key("libraries"): t.List(t.String),
}).ignore_extra("*")

machine_trafaret = t.Dict({
    t.Key("id"): t.String,
    t.Key("name"): t.String,
    t.Key("default"): t.Bool,
    t.Key("cpu"): t.String,
    t.Key("cpu_cores"): t.Int,
    t.Key("ephemeral_storage"): t.String,
    t.Key("has_gpu"): t.Bool,
    t.Key("memory"): t.String,
    t.Key("ram_gb"): t.Int,
}).ignore_extra("*")

notebook_execution_environment_trafaret = t.Dict({
    t.Key("image"): image_trafaret,
    t.Key("machine"): machine_trafaret,
    t.Key("time_to_live"): t.Int,
}).ignore_extra("*")


class ExecutionEnvironmentAssignPayload(TypedDict, total=False):
    """
    Payload for assigning an execution environment to a notebook.
    """

    environment_id: Optional[str]
    environment_slug: Optional[str]
    version_id: Optional[str]
    machine_id: Optional[str]
    machine_slug: Optional[str]
    time_to_live: int
    language: Optional[str]
    language_version: Optional[str]


class Image(APIObject):
    """
    Execution environment image information.

    Attributes
    ----------

    id : str
        The ID of the image.
    name : str
        The name of the image.
    default : bool
        Whether the image is the default image.
    description : str
        The description of the image.
    environment_id : str
        The ID of the environment.
    gpu_optimized : bool
        Whether the image is GPU optimized.
    language : ImageLanguage
        The runtime language of the image. For example "Python" or "R"
    language_version : str
        The version of the language. For example "3.11" or "4.3"
    libraries : list[str]
        A list of pre-installed libraries on the image.
    """

    _converter = image_trafaret

    def __init__(
        self,
        id: str,
        name: str,
        default: bool,
        description: str,
        environment_id: str,
        gpu_optimized: bool,
        language: ImageLanguage,
        language_version: str,
        libraries: List[str],
    ):
        self.id = id
        self.name = name
        self.default = default
        self.description = description
        self.environment_id = environment_id
        self.gpu_optimized = gpu_optimized
        self.language = language
        self.language_version = language_version
        self.libraries = libraries


class Machine(APIObject):
    """
    Execution environment machine information.

    Attributes
    ----------

    id : str
        The ID of the machine.
    name : str
        The name of the machine. Values include "XS", "S", "M", "L" etc.
    default : bool
        Whether the machine is the default machine.
    cpu : str
        The CPU of the machine. For example a value like "2000m".
    cpu_cores : int
        The number of CPU cores.
    ephemeral_storage : str
        The ephemeral storage of the machine. For example a value like "15Gi".
    has_gpu : bool
        Whether the machine has a GPU.
    memory : str
        The memory of the machine. For example a value like "8Gi".
    ram_gb : int
        The amount of RAM of the machine.
    """

    _converter = machine_trafaret

    def __init__(
        self,
        id: str,
        name: str,
        default: bool,
        cpu: str,
        cpu_cores: int,
        ephemeral_storage: str,
        has_gpu: bool,
        memory: str,
        ram_gb: int,
    ):
        self.id = id
        self.name = name
        self.default = default
        self.cpu = cpu
        self.cpu_cores = cpu_cores
        self.ephemeral_storage = ephemeral_storage
        self.has_gpu = has_gpu
        self.memory = memory
        self.ram_gb = ram_gb


class ExecutionEnvironment(APIObject):
    """
    An execution environment associated with a notebook.

    Attributes
    ----------

    image : Image
        The image associated with the execution environment.
    machine : Machine
        The machine associated with the execution environment.
    time_to_live : int
        The inactivity timeout for notebook session.
    """

    _path = "notebookExecutionEnvironments/"

    _converter = notebook_execution_environment_trafaret

    def __init__(
        self,
        image: Image,
        machine: Machine,
        time_to_live: int,
    ):
        self.image = image
        self.machine = machine
        self.time_to_live = time_to_live

    @classmethod
    def get(cls, notebook_id: str) -> ExecutionEnvironment:
        """
        Get a notebook execution environment by its notebook ID.

        Parameters
        ----------
        notebook_id : str
            The ID of the notebook.

        Returns
        -------
        ExecutionEnvironment
            The notebook execution environment.
        """
        r_data = cls._client.get(f"{cls._path}{notebook_id}/")
        json_data = r_data.json()
        return ExecutionEnvironment(
            image=Image.from_server_data(json_data["image"]),
            machine=Machine.from_server_data(json_data["machine"]),
            time_to_live=json_data["timeToLive"],
        )

    @classmethod
    def assign_environment(
        cls,
        notebook_id: str,
        payload: ExecutionEnvironmentAssignPayload,
    ) -> ExecutionEnvironment:
        """
        Assign execution environment values to a notebook.

        Parameters
        ----------
        notebook_id : str
            The ID of the notebook.
        payload : ExecutionEnvironmentAssignPayload
            The payload for the assignment/update.

        Returns
        -------
        ExecutionEnvironment
            The assigned execution environment.

        Examples
        --------
        .. code-block:: python

            from datarobot.models.notebooks import ExecutionEnvironment, ExecutionEnvironmentAssignPayload

            payload = ExecutionEnvironmentAssignPayload(machine_slug='medium', time_to_live=10)
            exec_env = ExecutionEnvironment.assign_environment('67914bfab0279fd832dc3fd1', payload)

        """
        r_data = cls._client.patch(f"{cls._path}{notebook_id}/", data=payload)
        json_data = r_data.json()
        return ExecutionEnvironment(
            image=Image.from_server_data(json_data["image"]),
            machine=Machine.from_server_data(json_data["machine"]),
            time_to_live=json_data["timeToLive"],
        )



#
# Copyright 2023-2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional, Union

import trafaret as t
from typing_extensions import TypedDict

from datarobot.enums import (
    CategoricalStatsMethods,
    DatetimeSamplingStrategy,
    DownsamplingOperations,
    FilterOperationFunctions,
    FindAndReplaceMatchMode,
    NumericStatsMethods,
    SamplingOperations,
    WranglingOperations,
    enum_to_list,
)
from datarobot.models.api_object import APIObject


class BaseOperation(APIObject):
    """Single base transformation unit in Data Wrangler recipe."""

    def __init__(self, directive: str, arguments: Any):
        self.directive = directive
        self.arguments = arguments


class WranglingOperation(BaseOperation):
    """Base class for data wrangling operations."""

    _converter = t.Dict({
        t.Key("directive"): t.Enum(*enum_to_list(WranglingOperations)),
        t.Key("arguments"): t.Mapping(t.String(), t.Any()),
    }).allow_extra("*")


class DownsamplingOperation(BaseOperation):
    """Base class for downsampling operations."""

    _converter = t.Dict({
        t.Key("directive"): t.Enum(*enum_to_list(DownsamplingOperations)),
        t.Key("arguments"): t.Mapping(t.String(), t.Any()),
    }).allow_extra("*")


class SamplingOperation(BaseOperation):
    _converter = t.Dict({
        t.Key("directive"): t.Enum(*enum_to_list(SamplingOperations)),
        t.Key("arguments"): t.Mapping(t.String(), t.Any()),
    }).allow_extra("*")


class BaseTimeAwareTask(APIObject):
    """Base class for time-aware tasks in time series operation task plan."""

    def __init__(self, name: str, arguments: Dict[str, Any]):
        self.name = name
        self.arguments = arguments


class TaskPlanElement(APIObject):
    """
    Represents a task plan element for a specific column in a time series operation.

    Parameters
    ----------
    column:
        Column name for which the task plan is defined.
    task_list:
        List of time-aware tasks to be applied to the column.
    """

    def __init__(self, column: str, task_list: List[BaseTimeAwareTask]):
        self.column = column
        self.task_list = task_list


class CategoricalStats(BaseTimeAwareTask):
    """
    Time-aware task to compute categorical statistics for a rolling window.

    Parameters
    ----------
    methods:
        List of categorical statistical methods to apply for rolling statistics.
    window_size:
        Number of rows to include in the rolling window.
    """

    def __init__(self, methods: List[CategoricalStatsMethods], window_size: int):
        super().__init__("categorical-stats", {"window_size": window_size, "methods": methods})


class NumericStats(BaseTimeAwareTask):
    """
    Time-aware task to compute numeric statistics for a rolling window.

    Parameters
    ----------
    methods:
        List of numeric statistical methods to apply for rolling statistics.
    window_size:
        Number of rows to include in the rolling window.
    """

    def __init__(self, methods: List[NumericStatsMethods], window_size: int):
        super().__init__("numeric-stats", {"window_size": window_size, "methods": methods})


class Lags(BaseTimeAwareTask):
    """
    Time-aware task to create one or more lags for a feature.

    Parameters
    ----------
    orders:
        List of lag orders to create.
    """

    def __init__(self, orders: List[int]):
        super().__init__("lags", {"orders": orders})


class LagsOperation(WranglingOperation):
    """
    Data wrangling operation to create one or more lags for a feature based off of a datetime ordering feature.
    This operation will create a new column for each lag order specified.

    Parameters
    ----------
    column:
        Column name to create lags for.
    orders:
        List of lag orders to create.
    datetime_partition_column:
        Column name used to partition the data by datetime. Used to order the data for lag creation.
    multiseries_id_column:
        Column name used to identify time series within the data. Required only for multiseries.

    Examples
    --------
    Create lags of orders 1, 5 and 30 in stock price data on opening price column "open_price", ordered by datetime
    column "date". The data contains multiple time series identified by "ticker_symbol":

    .. code-block:: python

        >>> import datarobot as dr
        >>> from datarobot.models.recipe_operation import LagsOperation
        >>> recipe = dr.Recipe.get('690bbf77aa31530d8287ae5f')
        >>> lags_op = LagsOperation(
        ...     column="open_price",
        ...     orders=[1, 5, 30],
        ...     datetime_partition_column="date",
        ...     multiseries_id_column="ticker_symbol",
        ... )
        >>> recipe.update(operations=[lags_op])
    """

    def __init__(
        self,
        column: str,
        orders: List[int],
        datetime_partition_column: str,
        multiseries_id_column: Optional[str] = None,
    ):
        super().__init__(
            directive=WranglingOperations.LAGS,
            arguments={
                "column": column,
                "orders": orders,
                "datetime_partition_column": datetime_partition_column,
                "multiseries_id_column": multiseries_id_column,
            },
        )


class WindowCategoricalStatsOperation(WranglingOperation):
    """
    Data wrangling operation to calculate categorical statistics for a rolling window. This operation
    will create a new column for each method specified.

    Parameters
    ----------
    column:
        Column name to create rolling statistics for.
    window_size:
        Number of rows to include in the rolling window.
    methods:
        List of methods to apply for rolling statistics. Currently only supports
        `datarobot.enums.CategoricalStatsMethods.MOST_FREQUENT`.
    datetime_partition_column:
        Column name used to partition the data by datetime. Used to order the timeseries data.
    multiseries_id_column:
        Column name used to identify each time series within the data. Required only for multiseries.
    rolling_most_frequent_udf:
        Fully qualified path to rolling most frequent user defined function. Used to optimize sql execution with
        snowflake.

    Examples
    --------
    Create rolling categorical statistics to track the most frequent product category purchased by customers based on
    their last 50 purchases:

    .. code-block:: python

        >>> import datarobot as dr
        >>> from datarobot.models.recipe_operation import WindowCategoricalStatsOperation
        >>> from datarobot.enums import CategoricalStatsMethods
        >>> recipe = dr.Recipe.get('690bbf77aa31530d8287ae5f')
        >>> window_cat_stats_op = WindowCategoricalStatsOperation(
        ...     column="product_category",
        ...     window_size=50,
        ...     methods=[CategoricalStatsMethods.MOST_FREQUENT],
        ...     datetime_partition_column="purchase_date",
        ...     multiseries_id_column="customer_id",
        ... )
        >>> recipe.update(operations=[window_cat_stats_op])
    """

    def __init__(
        self,
        column: str,
        window_size: int,
        methods: List[CategoricalStatsMethods],
        datetime_partition_column: str,
        multiseries_id_column: Optional[str] = None,
        rolling_most_frequent_udf: Optional[str] = None,
    ):
        super().__init__(
            directive=WranglingOperations.WINDOW_CATEGORICAL_STATS,
            arguments={
                "column": column,
                "window_size": window_size,
                "methods": methods,
                "datetime_partition_column": datetime_partition_column,
                "multiseries_id_column": multiseries_id_column,
                "rolling_most_frequent_user_defined_function": rolling_most_frequent_udf,
            },
        )


class WindowNumericStatsOperation(WranglingOperation):
    """
    Data wrangling operation to calculate numeric statistics for a rolling window. This operation will create one
    or more new columns.

    Parameters
    ----------
    column:
        Column name to create rolling statistics for.
    window_size:
        Number of rows to include in the rolling window.
    methods:
        List of methods to apply for rolling statistics. A new column will be created for each method.
    datetime_partition_column:
        Column name used to partition the data by datetime. Used to order the timeseries data.
    multiseries_id_column:
        Column name used to identify each time series within the data. Required only for multiseries.
    rolling_median_udf:
        Fully qualified path to a rolling median user-defined function. Used to optimize SQL execution with Snowflake.

    Examples
    --------
    Create rolling numeric statistics to track the maximum, minimum, and median stock prices over the last 7 trading
    sessions:

    .. code-block:: python

        >>> import datarobot as dr
        >>> from datarobot.models.recipe_operation import WindowNumericStatsOperation
        >>> from datarobot.enums import NumericStatsMethods
        >>> recipe = dr.Recipe.get('690bbf77aa31530d8287ae5f')
        >>> window_num_stats_op = WindowNumericStatsOperation(
        ...     column="stock_price",
        ...     window_size=7,
        ...     methods=[
        ...         NumericStatsMethods.MAX,
        ...         NumericStatsMethods.MIN,
        ...         NumericStatsMethods.MEDIAN,
        ...     ],
        ...     datetime_partition_column="trading_date",
        ...     multiseries_id_column="ticker_symbol",
        ... )
        >>> recipe.update(operations=[window_num_stats_op])
    """

    def __init__(
        self,
        column: str,
        window_size: int,
        methods: List[NumericStatsMethods],
        datetime_partition_column: str,
        multiseries_id_column: Optional[str] = None,
        rolling_median_udf: Optional[str] = None,
    ):
        super().__init__(
            directive=WranglingOperations.WINDOW_NUMERIC_STATS,
            arguments={
                "column": column,
                "window_size": window_size,
                "methods": methods,
                "datetime_partition_column": datetime_partition_column,
                "multiseries_id_column": multiseries_id_column,
                "rolling_median_user_defined_function": rolling_median_udf,
            },
        )


class TimeSeriesOperation(WranglingOperation):
    """
    Data wrangling operation to generate a dataset ready for time series modeling: with forecast point, forecast
    distances, known in advance columns, etc.

    Parameters
    ----------
    target_column:
        Target column to use for generating naive baseline features during feature reduction.
    datetime_partition_column:
        Column name used to partition the data by datetime. Used to order the time series data.
    forecast_distances:
        List of forecast distances to generate features for. Each distance represents a relative position that
        determines how many rows ahead to predict.
    task_plan:
        List of task plans for each column.
    baseline_periods:
        List of integers representing the periodicities used to generate naive baseline features from the target.
        Baseline period = 1 corresponds to the naive latest baseline.
    known_in_advance_columns:
        List of columns that are known in advance at prediction time, i.e. features that do not need to be lagged.
    multiseries_id_column:
        Column name used to identify each time series within the data. Required only for multiseries.
    rolling_median_udf:
        Fully qualified path to rolling median user defined function. Used to optimize SQL execution with Snowflake.
    rolling_most_frequent_udf:
        Fully qualified path to rolling most frequent user defined function.
    forecast_point:
        To use at prediction time.

    Examples
    --------
    Create a time series operation for sales forecasting with forecast distances of 7 and 30 days, using the sale amount
    as the target column, the date of the sale for datetime ordering, and "store_id" as the multiseries identifier.
    The operation includes a task plan to compute lags of orders 1, 7, and 30 on the sales amount, and
    specifies known in advance columns "promotion" and "holiday_flag":

    .. code-block:: python

        >>> import datarobot as dr
        >>> from datarobot.models.recipe_operation import TimeSeriesOperation, TaskPlanElement, Lags
        >>> recipe = dr.Recipe.get('690bbf77aa31530d8287ae5f')
        >>> task_plan = [
        ...     TaskPlanElement(
        ...         column="sales_amount",
        ...         task_list=[Lags(orders=[1, 7, 30])]
        ...     )
        ... ]
        >>> time_series_op = TimeSeriesOperation(
        ...     target_column="sales_amount",
        ...     datetime_partition_column="sale_date",
        ...     forecast_distances=[7, 30],
        ...     task_plan=[task_plan],
        ...     known_in_advance_columns=["promotion", "holiday_flag"],
        ...     multiseries_id_column="store_id"
        ... )
        >>> recipe.update(operations=[time_series_op])
    """

    def __init__(
        self,
        target_column: str,
        datetime_partition_column: str,
        forecast_distances: List[int],
        task_plan: List[TaskPlanElement],
        baseline_periods: Optional[List[int]] = None,
        known_in_advance_columns: Optional[List[str]] = None,
        multiseries_id_column: Optional[str] = None,
        rolling_median_udf: Optional[str] = None,
        rolling_most_frequent_udf: Optional[str] = None,
        forecast_point: Optional[datetime] = None,
    ):
        arguments = {
            "target_column": target_column,
            "datetime_partition_column": datetime_partition_column,
            "forecast_distances": forecast_distances,
            "task_plan": task_plan,
            "multiseries_id_column": multiseries_id_column,
            "known_in_advance_columns": known_in_advance_columns,
            "baseline_periods": baseline_periods,
            "rolling_median_user_defined_function": rolling_median_udf,
            "rolling_most_frequent_user_defined_function": rolling_most_frequent_udf,
            "forecast_point": forecast_point,
        }
        super().__init__(directive=WranglingOperations.TIME_SERIES, arguments=arguments)


class ComputeNewOperation(WranglingOperation):
    """
    Data wrangling operation to create a new feature computed using a SQL expression.

    Parameters
    ----------
    expression:
        SQL expression to compute the new feature.
    new_feature_name:
        Name of the new feature.

    Examples
    --------
    Create a new feature "total_sales" by summing the total of "online_sales" and "in_store_sales", rounded to the
    nearest dollar:

    .. code-block:: python

        >>> import datarobot as dr
        >>> from datarobot.models.recipe_operation import ComputeNewOperation
        >>> recipe = dr.Recipe.get('690bbf77aa31530d8287ae5f')
        >>> compute_new_op = ComputeNewOperation(
        ...     expression="ROUND(online_sales + in_store_sales, 0)",
        ...     new_feature_name="total_sales"
        ... )
        >>> recipe.update(operations=[compute_new_op])
    """

    def __init__(self, expression: str, new_feature_name: str):
        super().__init__(
            directive=WranglingOperations.COMPUTE_NEW,
            arguments={"expression": expression, "new_feature_name": new_feature_name},
        )


class RenameColumnsOperation(WranglingOperation):
    """
    Data wrangling operation to rename one or more columns.

    Parameters
    ----------
    column_mappings:
        Mapping of original column names to new column names.

    Examples
    --------
    Rename columns "old_name1" to "new_name1" and "old_name2" to "new_name2":

    .. code-block:: python

        >>> import datarobot as dr
        >>> from datarobot.models.recipe_operation import RenameColumnsOperation
        >>> recipe = dr.Recipe.get('690bbf77aa31530d8287ae5f')
        >>> rename_op = RenameColumnsOperation(
        ...     column_mappings={'old_name1': 'new_name1', 'old_name2': 'new_name2'}
        ... )
        >>> recipe.update(operations=[rename_op])
    """

    def __init__(self, column_mappings: Dict[str, str]):
        """
        column_mapping: dict, where
            key:  str
                Original name
            value: str
                New name
        """
        super().__init__(
            directive=WranglingOperations.RENAME_COLUMNS,
            arguments={"column_mappings": [{"original_name": k, "new_name": v} for k, v in column_mappings.items()]},
        )


class FilterCondition(TypedDict):
    """
    Condition to filter rows in a FilterOperation.

    Parameters
    ----------
    column: str
        Column name to apply the condition on.
    function: :class:`FilterOperationFunctions <datarobot.enums.FilterOperationFunctions>`
        The filtering function to use.
    function_arguments: List[Union[str, int, float]]
        The list of arguments for the filtering function.

    Examples
    --------
    FilterCondition to filter rows where "age" is between 18 and 65:

    .. code-block:: python

        >>> from datarobot.models.recipe_operation import FilterCondition
        >>> from datarobot.enums import FilterOperationFunctions
        >>> condition = FilterCondition(
        ...     column="age",
        ...     function=FilterOperationFunctions.BETWEEN,
        ...     function_arguments=[18, 65]
        ... )
    """

    column: str
    function: FilterOperationFunctions
    function_arguments: List[Union[str, int, float]]


class FilterOperation(WranglingOperation):
    """
    Data wrangling operation to filter rows based on one or more conditions.

    Parameters
    ----------
    conditions:
        List of conditions to filter on.
    keep_rows:
        If matching rows should be kept or dropped.
    operator:
        Operator to use between conditions when using multiple conditions. Allowed values: [and, or].

    Examples
    --------
    Filter input to only keep users older than 18:

    .. code-block:: python

        >>> import datarobot as dr
        >>> from datarobot.models.recipe_operation import FilterOperation, FilterCondition
        >>> from datarobot.enums import FilterOperationFunctions
        >>> recipe = dr.Recipe.get('690bbf77aa31530d8287ae5f')
        >>> condition = FilterCondition(
        ...     column="age",
        ...     function=FilterOperationFunctions.GREATER_THAN,
        ...     function_arguments=[18]
        ... )
        >>> filter_op = FilterOperation(conditions=[condition], keep_rows=True)
        >>> recipe.update(operations=[filter_op])

    Filter input to filter out rows where "status" is either "inactive" or "banned":

    .. code-block:: python

        >>> import datarobot as dr
        >>> from datarobot.models.recipe_operation import FilterOperation, FilterCondition
        >>> from datarobot.enums import FilterOperationFunctions
        >>> recipe = dr.Recipe.get('690bbf77aa31530d8287ae5f')
        >>> inactive_cond = FilterCondition(
        ...     column="status",
        ...     function=FilterOperationFunctions.EQUALS,
        ...     function_arguments=["inactive"]
        ... )
        >>> banned_cond = FilterCondition(
        ...     column="status",
        ...     function=FilterOperationFunctions.EQUALS,
        ...     function_arguments=["banned"]
        ... )
        >>> filter_op = FilterOperation(
        ...     conditions=[inactive_cond, banned_cond],
        ...     keep_rows=False,
        ...     operator="or"
        ... )
        >>> recipe.update(operations=[filter_op])
    """

    def __init__(
        self,
        conditions: List[FilterCondition],
        keep_rows: Optional[bool] = True,
        operator: Optional[str] = "and",
    ):
        super().__init__(
            directive=WranglingOperations.FILTER,
            arguments={"keep_rows": keep_rows, "operator": operator, "conditions": conditions},
        )


class DropColumnsOperation(WranglingOperation):
    """
    Data wrangling operation to drop one or more columns.

    Parameters
    ----------
    columns:
        Columns to drop.

    Examples
    --------
    .. code-block:: python

        >>> import datarobot as dr
        >>> from datarobot.models.recipe_operation import DropColumnsOperation
        >>> recipe = dr.Recipe.get('690bbf77aa31530d8287ae5f')
        >>> drop_op = DropColumnsOperation(columns=['col1', 'col2'])
        >>> recipe.update(operations=[drop_op])
    """

    def __init__(self, columns: List[str]):
        super().__init__(
            directive=WranglingOperations.DROP_COLUMNS,
            arguments={"columns": columns},
        )


class DedupeRowsOperation(WranglingOperation):
    """
    Data wrangling operation to remove duplicate rows. Uses values from all columns.

    Examples
    --------
    .. code-block:: python

        >>> import datarobot as dr
        >>> from datarobot.models.recipe_operation import DedupeRowsOperation
        >>> recipe = dr.Recipe.get('690bbf77aa31530d8287ae5f')
        >>> dedupe_op = DedupeRowsOperation()
        >>> recipe.update(operations=[dedupe_op])
    """

    def __init__(self):  # type: ignore[no-untyped-def]
        super().__init__(
            directive=WranglingOperations.DEDUPE_ROWS,
            arguments={},
        )


class FindAndReplaceOperation(WranglingOperation):
    """
    Data wrangling operation to find and replace strings in a column.

    Parameters
    ----------
    column:
        Column name to perform find and replace on.
    find:
        String or expression to find.
    replace_with:
        String to replace with.
    match_mode:
        Match mode to use when finding strings.
    is_case_sensitive:
        Whether the find operation should be case sensitive.

    Examples
    --------
    Set Recipe operations to search for exact match of "old_value" in column "col1" and replace with "new_value":

    .. code-block:: python

        >>> import datarobot as dr
        >>> from datarobot.models.recipe_operation import FindAndReplaceOperation
        >>> from datarobot.enums importFindAndReplaceMatchMode
        >>> recipe = dr.Recipe.get('690bbf77aa31530d8287ae5f')
        >>> find_replace_op = FindAndReplaceOperation(
        ...     column="col1",
        ...     find="old_value",
        ...     replace_with="new_value",
        ...     match_mode=FindAndReplaceMatchMode.EXACT,
        ...     is_case_sensitive=True
        ... )
        >>> recipe.update(operations=[find_replace_op])

    Set Recipe operations to use regular expression to replace names starting with "Brand" in column "name" and replace
    with "Lyra":

    .. code-block:: python

        >>> import datarobot as dr
        >>> from datarobot.models.recipe_operation import FindAndReplaceOperation
        >>> from datarobot.enums import FindAndReplaceMatchMode
        >>> recipe = dr.Recipe.get('690bbf77aa31530d8287ae5f')
        >>> find_replace_op = FindAndReplaceOperation(
        ...     column="name",
        ...     find="^Brand.*",
        ...     replace_with="Lyra",
        ...     match_mode=FindAndReplaceMatchMode.REGEX
        ... )
        >>> recipe.update(operations=[find_replace_op])
    """

    def __init__(
        self,
        column: str,
        find: str,
        replace_with: str,
        match_mode: FindAndReplaceMatchMode = FindAndReplaceMatchMode.EXACT,
        is_case_sensitive: bool = False,
    ):
        super().__init__(
            directive=WranglingOperations.REPLACE,
            arguments={
                "origin": column,
                "searchFor": find,
                "replacement": replace_with,
                "matchMode": match_mode.value,
                "isCaseSensitive": is_case_sensitive,
            },
        )


class RandomSamplingOperation(SamplingOperation):
    def __init__(self, rows: int, seed: Optional[int] = None):
        super().__init__(
            directive=SamplingOperations.RANDOM_SAMPLE,
            arguments={"rows": rows, "seed": seed},
        )


class DatetimeSamplingOperation(SamplingOperation):
    def __init__(
        self,
        datetime_partition_column: str,
        rows: int,
        strategy: Optional[Union[str, DatetimeSamplingStrategy]] = None,
        multiseries_id_column: Optional[str] = None,
        selected_series: Optional[List[str]] = None,
    ):
        super().__init__(
            directive=SamplingOperations.DATETIME_SAMPLE,
            arguments={
                "rows": rows,
                "strategy": strategy,
                "datetime_partition_column": datetime_partition_column,
                "multiseries_id_column": multiseries_id_column,
                "selected_series": selected_series,
            },
        )


class RandomDownsamplingOperation(DownsamplingOperation):
    """
    A downsampling technique that reduces the size of the majority class using random sampling (i.e., each sample has an
    equal probability of being chosen).

    Parameters
    ----------
    max_rows:
        The maximum number of rows to downsample to.
    seed:
        The random seed to use for downsampling. Optional.

    Examples
    --------
    .. code-block:: python

        >>> from datarobot.models.recipe_operation import RandomDownsamplingOperation
        >>> op = RandomDownsamplingOperation(max_rows=600)
    """

    DEFAULT_SEED = 43  # Default seed used in UI

    def __init__(self, max_rows: int, seed: int = DEFAULT_SEED):
        super().__init__(
            directive=DownsamplingOperations.RANDOM_SAMPLE,
            arguments={"rows": max_rows, "seed": seed},
        )


class SmartDownsamplingMethod(Enum):
    """Smart downsampling methods."""

    BINARY = "binary"
    ZERO_INFLATED = "zero-inflated"


class SmartDownsamplingOperation(DownsamplingOperation):
    """
    A downsampling technique that relies on the distribution of target values to adjust size and specifies how much a
    specific class was sampled in a new column.

    For this technique to work, ensure the recipe has set `target` and `weightsFeature` in the recipe's settings.

    Parameters
    ----------
    max_rows:
        The maximum number of rows to downsample to.
    method:
        The downsampling method to use.
    seed:
        The random seed to use for downsampling. Optional.

    Examples
    --------
    .. code-block:: python

        >>> from datarobot.models.recipe_operation import SmartDownsamplingOperation, SmartDownsamplingMethod
        >>> op = SmartDownsamplingOperation(max_rows=1000, method=SmartDownsamplingMethod.BINARY)
    """

    DEFAULT_SEED = 43  # Default seed used in UI

    def __init__(
        self,
        max_rows: int,
        method: SmartDownsamplingMethod = SmartDownsamplingMethod.BINARY,
        seed: int = DEFAULT_SEED,
    ):
        super().__init__(
            directive=DownsamplingOperations.SMART_DOWNSAMPLING,
            arguments={"rows": max_rows, "method": method.value, "seed": seed},
        )
